# Copyright (C) 2016 Jeandre Kruger
#
# This file is part of s2scm.
#
# s2scm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# s2scm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with s2scm.  If not, see <http://www.gnu.org/licenses/>.

import kurt
from procedures import procedures, init_builtins
from reader import read

class ArgumentsError(Exception):
    pass

# Script new blocks should be appended to.
current_script = None

# Current environment.
current_env = []

# Number of continuations.
cont_count = 0

def get_env_vals(env):
    return map(lambda x: x[1], env)

def get_env_inserts(env):
    return map(lambda x: x[0], env)

def continuation_type(env):
    global cont_count
    t = kurt.CustomBlockType('stack', ['continuation_%d' % cont_count] + get_env_inserts(env))
    cont_count += 1
    return t

def compile_proc_call(call, sprite, script):
    """Compile the procedure call call."""
    script.append(kurt.Block(procedures[call[0].str][0],
                             *map(lambda x: compile_sexpr(x, sprite), call[1:])))
    script.append(kurt.Block(procedures['get return'][0]))

    return kurt.Block('readVariable', 'return')

def special_form_define(sexpr, sprite, script = None):
    global current_script

    if len(sexpr) > 3:
        raise ArgumentsError('Too many arguments for define')
    elif len(sexpr) < 3:
        raise ArgumentsError('Not enough arguments for define')

    if script is None:
        script = current_script

    name = sexpr[1].str
    val = compile_sexpr(sexpr[2], sprite, script)
    current_env.append((kurt.Insert('string', name = name), val))
    cont_type = continuation_type(current_env)
    cont_call = kurt.Block(cont_type, *get_env_vals(current_env))
    cont = [kurt.Block('procDef', cont_type)]

    script.append(cont_call)

    sprite.scripts.append(kurt.Script(cont))

    # Start adding new blocks at the continuation.
    current_script = cont

    return kurt.Block('getParam', name)

def special_form_if(sexpr, sprite, script = None):
    if len(sexpr) > 4:
        raise ArgumentsError('Too many arguments for if')
    elif len(sexpr) < 3:
        raise ArgumentsError('Not enough arguments for if')

    if script is None:
        script = current_script

    if len(sexpr) == 3:
        then_clause = []
        ret = compile_sexpr(sexpr[2], sprite, then_clause)
        then_clause.append(kurt.Block('setVar:to:', 'ifVal', ret))
        script.append(kurt.Block('doIf', compile_sexpr(sexpr[1], sprite),
                                 then_clause))
    else:
        then_clause = []
        else_clause = []
        then_ret = compile_sexpr(sexpr[2], sprite, then_clause)
        else_ret = compile_sexpr(sexpr[3], sprite, else_clause)
        then_clause.append(kurt.Block('setVar:to:', 'ifVal', then_ret))
        else_clause.append(kurt.Block('setVar:to:', 'ifVal', else_ret))
        script.append(kurt.Block('doIfElse', compile_sexpr(sexpr[1], sprite),
                                 then_clause, else_clause))

    return kurt.Block('readVariable', 'ifVal')

special_forms = {
    'define': special_form_define,
    'if':     special_form_if
}

def compile_sexpr(sexpr, sprite, script = None):
    """Compile a single S-Expression and return a list of kurt.Block."""
    if script is None:
        script = current_script

    if isinstance(sexpr, list) and sexpr != []:
        if sexpr[0].str in special_forms:
            return special_forms[sexpr[0].str](sexpr, sprite, script)
        else:
            return compile_proc_call(sexpr, sprite, script)
    else:
        return sexpr

def compile_project(infile, outfile):
    """Compile Scheme to Scratch.

    Take the Scheme file infile as input and outfile as output."""

    global current_script

    project = kurt.Project()

    sprite = kurt.Sprite(project, 'Sprite')
    sprite.scripts.append(kurt.Script([kurt.Block("whenGreenFlag")]))
    current_script = sprite.scripts[0]
    init_builtins(sprite)

    f = open(infile, 'r')

    while True:
        sexpr = read(f)
        if sexpr == '':
            break
        compile_sexpr(sexpr, sprite)

    f.close()

    project.sprites.append(sprite)

    project.save(outfile)



# Copyright (C) 2016 Jeandre Kruger
#
# This file is part of s2scm.
#
# s2scm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# s2scm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with s2scm.  If not, see <http://www.gnu.org/licenses/>.

import sys, re

class Symbol:
    """Represents a symbol."""
    def __init__(self, s):
        self.str = s.lower()

    def __repr__(self):
        return "Symbol(%r)" % self.str

def is_symbol_char(char, first = False):
    if first:
        return bool(re.match(r'[a-zA-Z!$%&*+./:<=>?@^_~-]', char))
    return bool(re.match(r'[a-zA-Z0-9!$%&*+./:<=>?@^_~-]', char))

def read_symbol(char, f = sys.stdin):
    """Read a symbol from the file f.

    Start with the character char."""
    s = ''

    while True:
        s += char

        next_char = f.read(1)

        if is_symbol_char(next_char):
            char = next_char
        else:
            # Unread the character.
            f.seek(-1, 1)
            return Symbol(s)

def read_num(char, f = sys.stdin):
    """Read a number from the file f.

    Start with the character char."""
    s = ''

    read_decimal = False

    while True:
        s += char

        next_char = f.read(1)

        if re.match(r'[0-9]', next_char):
            char = next_char
        elif re.match(r'[0-9.]', next_char) and not read_decimal:
            char = next_char
            read_decimal = True
        else:
            # Unread the character.
            f.seek(-1, 1)

            if read_decimal:
                return float(s)
            else:
                return int(s)

def read_string(f = sys.stdin):
    """Read a string from the file f."""
    # Escape sequences
    escapes = {
        'a': '\a',
        'b': '\b',
        'f': '\f',
        'n': '\n',
        'r': '\r',
        't': '\t',
        'v': '\v'
    }

    s = ''

    while True:
        char = f.read(1)

        if char == '':
            raise SyntaxError('Unexpected EOF, expected "')

        if char == '"':
            break

        if char == '\\':
            # Escape sequence
            escape = f.read(1)

            if escape == '':
                raise SyntaxError('Unexpected EOF after \\ in string')

            if escape in escapes:
                s += escapes[escape]
            else:
                s += escape
        else:
            s += char

    return s

def read_hash(f = sys.stdin):
    """Read a boolean or character from the file f."""
    char = f.read(1)

    if char == '':
        raise SyntaxError('Unexpected EOF after #')

    if char == 't':
        return True
    elif char == 'f':
        return False
    elif char == '\\':
        # Character
        ch = f.read(1)
        if ch == '':
            raise SyntaxError('Unexpected EOF after #\\')
        return ch
    else:
        raise SyntaxError('Unexpected #%s' % char)

def read_token(f = sys.stdin):
    """Read a single token from the file f."""
    char = f.read(1)

    if char == '':
        return ''
    elif re.match(r'[\n\t\v ]', char):
        return read_token(f)
    elif char == '(' or char == ')':
        return char
    elif char == '"':
        return read_string(f)
    elif char == '#':
        return read_hash(f)
    elif is_symbol_char(char, True):
        return read_symbol(char, f)
    elif re.match(r'[0-9]', char):
        return read_num(char, f)
    else:
        raise SyntaxError('Unexpected character %s' % char)

def read_list(f = sys.stdin):
    """Read a list from the file f."""
    tok = read_token(f)

    if tok == '':
        raise SyntaxError('Unexpected EOF, expected )')
    elif tok == ')':
        return []
    elif tok == '(':
        lst = read_list(f)
        return [lst] + read_list(f)
    else:
        return [tok] + read_list(f)

def read(f = sys.stdin):
    """Read a single S-Expression from the file f."""
    tok = read_token(f)

    if tok == '(':
        return read_list(f)
    elif tok == ')':
        raise SyntaxError('Unexpected )')
    else:
        return tok



# Copyright (C) 2016 Jeandre Kruger
#
# This file is part of s2scm.
#
# s2scm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# s2scm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with s2scm.  If not, see <http://www.gnu.org/licenses/>.

from distutils.core import setup

setup(
    name = "s2scm",
    version = "0.2.4",
    description = "Compile Scheme to MIT Scratch",
    author = "Jeandre Kruger",
    author_email = "Unknown",
    url = "https://github.com/Jonathan50/s2scm",
    packages = [ "s2scm" ],
    install_requires = [ "kurt" ]
)



# Copyright (C) 2016 Jeandre Kruger
#
# This file is part of s2scm.
#
# s2scm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# s2scm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with s2scm.  If not, see <http://www.gnu.org/licenses/>.

import kurt

def arg(t, name):
    """Get a custom block argument of type t and name name."""
    shapes = {
        'n': 'number',
        's': 'string'
    }
    return kurt.Insert(shapes[t], name = name)

def proc_return(val = ''):
    return kurt.Block('append:toList:', val, 'return stack')

def proc_get_return(blocktype):
    return [kurt.Block('setVar:to:', 'return', kurt.Block('getLine:ofList:', 'last', 'return stack')),
            kurt.Block('deleteLine:ofList:', 'last', 'return stack')]

def proc_add(blocktype):
    return [proc_return(kurt.Block('+', kurt.Block('getParam', 'x'), kurt.Block('getParam', 'y')))]
def proc_sub(blocktype):
    return [proc_return(kurt.Block('-', kurt.Block('getParam', 'x'), kurt.Block('getParam', 'y')))]
def proc_mul(blocktype):
    return [proc_return(kurt.Block('*', kurt.Block('getParam', 'x'), kurt.Block('getParam', 'y')))]
def proc_div(blocktype):
    return [proc_return(kurt.Block('/', kurt.Block('getParam', 'x'), kurt.Block('getParam', 'y')))]

def proc_say(blocktype):
    return [kurt.Block('say:duration:elapsed:from:',
                       kurt.Block('getParam', 's'), kurt.Block('getParam', 'duration')),
            proc_return()]

# All the builtin procedures
procedures = {
    'get return': (kurt.CustomBlockType('stack', ['get return']), proc_get_return),
    '+':          (kurt.CustomBlockType('stack', ['+', arg('n', 'x'), arg('n', 'y')]), proc_add),
    '-':          (kurt.CustomBlockType('stack', ['-', arg('n', 'x'), arg('n', 'y')]), proc_sub),
    '*':          (kurt.CustomBlockType('stack', ['*', arg('n', 'x'), arg('n', 'y')]), proc_mul),
    '/':          (kurt.CustomBlockType('stack', ['/', arg('n', 'x'), arg('n', 'y')]), proc_div),
    'say':        (kurt.CustomBlockType('stack', ['say', arg('s', 's'), arg('n', 'duration')]), proc_say)
}

def init_builtins(sprite):
    """Add all the builtin procedures to the sprite sprite."""
    for proc in procedures:
        script = [kurt.Block('procDef', procedures[proc][0])] + procedures[proc][1](procedures[proc][0])
        sprite.scripts.append(kurt.Script(script))



import sys
from project import *

def usage(status = 0):
    print 'Usage: s2scm infile.scm outfile.sb2'
    sys.exit(status)

if '-h' in sys.argv or '-?' in sys.argv or '--help' in sys.argv:
    usage()
elif '-v' in sys.argv or '--version' in sys.argv:
    print '''s2scm version 0.2.4

Copyright (C) 2016 Jeandre Kruger.

s2scm is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

s2scm is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with s2scm.  If not, see <http://www.gnu.org/licenses/>.'''
    sys.exit(0)

if len(sys.argv) != 3:
    usage(1)

compile_project(sys.argv[1], sys.argv[2])
